import java.util.*;
import java.io.*;

public class MaxCliqueBB {
    private int[] degree;
    private int[][] A;          // adjacency matrix
    private int n,m;            // n vertices, m edges
    private long timeLimit;     // seconds
    private long cpuTime;       // milliseconds
    private long nodes;
    private int[] solution;
    private int maxSize;
    private int style;          // style of static ordering
    private boolean trace;
    private BitSet[] N;         // neighbourhood
    private BitSet[] notN;      // not the neighbourhood
    private Vertex[] V;         // vertices sorted by degree
    private int maxColVert;     // vertex of maximum colour, byproduct of BBCol

    public MaxCliqueBB (String fname) throws IOException {
	String b     = "";
	Scanner sc   = new Scanner(new File(fname));
	while (sc.hasNext() && !b.equals("p")) b = sc.next();
	sc.next();
	n          = sc.nextInt();          // # vertices
	m          = sc.nextInt();          // # edges
	degree     = new int[n];
	A          = new int[n][n];
	N          = new BitSet[n];
	notN       = new BitSet[n];
	solution   = new int[n];
	V          = new Vertex[n];
	for (int i=0;i<n;i++){	    
	    degree[i] = 0;
	    Arrays.fill(A[i],0);
	    N[i] = new BitSet(n);
	    notN[i] = new BitSet(n);
	}
	while (sc.hasNext()){
	    b     = sc.next(); // skip "edge"
	    int i = sc.nextInt() - 1;
	    int j = sc.nextInt() - 1; 
	    degree[i]++; degree[j]++;
	    A[i][j] = A[j][i] = 1;
	}
	sc.close();
	for (int i=0;i<n;i++) V[i] = new Vertex(i,degree[i]);
	nodes = maxSize = 0;
	timeLimit = -1;
	cpuTime = -1;
	style = 1;
	trace = false;
	maxColVert = -1;
    }

    private void minWidthOrder(){
	ArrayList<Vertex> L = new ArrayList<Vertex>(n);
	Stack<Vertex> S     = new Stack<Vertex>();
	for (Vertex v : V){v.init(n); L.add(v);}
	while (!L.isEmpty()){
	    Vertex v = L.get(0);
	    for (Vertex u : L) if (u.getDegree() < v.getDegree()) v = u;
	    S.push(v);
	    L.remove(v);
	    for (Vertex u : L) if (A[u.getIndex()][v.getIndex()] == 1) u.decrementDegree();
	}
	int k=0;
	while (!S.isEmpty()){V[k] = S.pop(); k++;}
    }
	

    private void brelazOrder(){
	ArrayList<Vertex> L = new ArrayList<Vertex>(n);
	ArrayList<Vertex> ordered = new ArrayList<Vertex>(n);
	for (Vertex v : V){v.init(n); L.add(v);}
	while (!L.isEmpty()){
	    Vertex v = selectMaxSaturatedUncolouredVertex(L);
	    colourVertex(v);
	    L.remove(v);
	    ordered.add(v);
	}
	int k=0;
	for (Vertex v : ordered){V[k] = v; k++;};
    }

    private Vertex selectMaxSaturatedUncolouredVertex(ArrayList<Vertex> L){
	Vertex v = L.get(0);
	for (Vertex w : L)
	    if (w.getSaturation() > v.getSaturation() ||
		w.getSaturation() == v.getSaturation() && w.getDegree() > v.getDegree())
		v = w;
	return v;
    }

    private void colourVertex(Vertex v){
	int i   = v.getIndex();
	int col = v.colour();
	for (int j=0;j<n;j++)
	    if (A[i][j] == 1 && V[j].hasColour(col)){
		V[j].removeColour(col);
		V[j].incrementSaturation();
	    }
    }

    private void renameVertices(){
        if (style == 1) Arrays.sort(V,new DegreeComparator()); // degree order
	if (style == 2) minWidthOrder();                       // min width (smallest-last) order
	if (style == 3) brelazOrder();                         // brelaz order
	if (style == 4) {                                      // saturation order
	    brelazOrder();
	    Arrays.sort(V,new SaturationComparator()); 
	}
	if (trace) for (Vertex v : V) System.out.println(v);
	for (int i=0;i<n;i++)
	    for (int j=0;j<n;j++){
		int u = V[i].getIndex();
		int v = V[j].getIndex();
		N[i].set(j,A[u][v] == 1);
		notN[i].set(j,A[u][v] == 0);
	    }
    }
    
    public long getNodes(){return nodes;}
    public int  getMaxSize(){return maxSize;}
    public void setTrace(){trace = true;}
    public void setTimeLimit(long seconds){timeLimit = seconds;}
    public void setStyle(int style){this.style = style;}
    
    public void search(){
	nodes    = 0;
	renameVertices();
	cpuTime  = System.currentTimeMillis();
	BitSet C = new BitSet(n);
	BitSet P = new BitSet(n);
	P.set(0,n,true);
	greedyBuild(C,P);
	maxClique(C,P);
    }
    
    private void maxClique(BitSet C,BitSet P){
	if (timeLimit > 0 && (System.currentTimeMillis() - cpuTime)/1000 >= timeLimit) return;
	nodes++;
	int Pcard = P.cardinality();
	int[] U   = new int[Pcard];
	int[] Col = new int[Pcard];
	BBColour(C,(BitSet)P.clone(),U,Col);
	//numberSort(C,(BitSet)P.clone(),U,Col);
	for (int i=Pcard-1;i>=0 && Col[i] + C.cardinality() > maxSize;i--){
	    BitSet C_new = (BitSet)C.clone(); 
	    BitSet P_new = (BitSet)P.clone();	    
	    int v        = U[i];
	    select(v,C_new,P_new);
	    maxClique(C_new,P_new);
	    reject(v,C,P);
	}
    }

    private void select(int v,BitSet C,BitSet P){
	C.set(v,true);
	P.and(N[v]);
	if (C.cardinality() > maxSize) saveSolution(C);
    }

    private void reject(int v,BitSet C,BitSet P){
	P.set(v,false);
    }
    
    private int chooseMaxDegree(BitSet P){return P.nextSetBit(0);}
    
    private int chooseMinDegree(BitSet P){
	for (int i=n-1;i>=0;i--) if (P.get(i)) return i;
	return -1;
    }
    
    private int getDegree(int v,BitSet P){ // get active degree of vertex v in P
	BitSet P_new = (BitSet)P.clone();
	P_new.and(N[v]);
	return P_new.cardinality();
    }
    
    private int BBColour(BitSet C,BitSet P,int[] U,int[] Col){
	int colour = 0;
	int i = 0;
       	while (P.cardinality() != 0){
	    colour++;
	    BitSet Q = (BitSet)P.clone();
	    while (Q.cardinality() != 0){
		int v = chooseMaxDegree(Q);
		P.set(v,false);
		Q.set(v,false);
		Q.and(notN[v]);
		U[i] = v;        // add vertices to U in colour class order
		Col[i] = colour; // and its respective colour
		i++;
	    }
	}
	return colour;           // maximum colours required, possibly zero!
    }
    
    private boolean compatible(BitSet C,int v){
	BitSet C_clone = (BitSet)C.clone();
	C_clone.and(N[v]);
	return C_clone.isEmpty();
    }

    private int numberSort(BitSet C,BitSet P,int[] U,int[] Col){
	int colours = 0;
	BitSet [] ColourClass = new BitSet[P.cardinality()];
	//for (int i=0;i<P.cardinality();i++) ColourClass[i] = new BitSet(P.cardinality());
	for (int v = P.nextSetBit(0);v >= 0;v = P.nextSetBit(v+1)){
	    int k = 0;
	    while (ColourClass[k] != null && !compatible(ColourClass[k],v)) k++;
	    if (ColourClass[k] == null) ColourClass[k] = new BitSet(P.cardinality());
	    ColourClass[k].set(v,true);
	    colours = Math.max(colours,k+1);
	}
	int i=0;
	for (int k=0;k<colours;k++){
	    for (int j = ColourClass[k].nextSetBit(0);j >= 0;j = ColourClass[k].nextSetBit(j+1)){
		U[i] = j;
		Col[i] = k+1;
		i++;
	    }
	}
	return colours;
    }
    //
    // Tomita & Seki DMTCS 2003
    //	    
    
    private void saveSolution(BitSet C){
	Arrays.fill(solution,0);
	for (int i=0;i<n;i++) if (C.get(i)) solution[V[i].getIndex()] = 1;
	maxSize = C.cardinality();
    }
    
    public String getSolution(){
	String s = "";
	for (int i=0;i<n;i++)
	    if (solution[i] == 1) s = s + i + " ";
	return s;
    }

    private void greedyBuild(BitSet C,BitSet P){
	BitSet C_new = new BitSet(n);
	BitSet P_new = (BitSet)P.clone();
	while (P_new.cardinality() > 0) select(chooseMaxDegree(P_new),C_new,P_new);
	saveSolution(C_new);
	for (int i = P.nextSetBit(0);i >= 0;i = P.nextSetBit(i+1)){
	    degree[i] = getDegree(i,P);
	    if (degree[i] < maxSize) P.set(i,false);
	}
    }
    
    public static void main(String[] args)  throws IOException {
	if (args.length == 0){
	    System.out.println("fname cpuLimit style trace");
	    return;
	}
	String fname   = args[0];
	MaxCliqueBB mc = new MaxCliqueBB(fname);
	if (args.length > 1) mc.setTimeLimit((long)Integer.parseInt(args[1]));
	if (args.length > 2) mc.setStyle(Integer.parseInt(args[2]));
	//if (args.length > 3) mc.setTrace();
	long cpuTime = System.currentTimeMillis();
	mc.search();
	cpuTime = System.currentTimeMillis() - cpuTime;
	System.out.println(mc.getMaxSize() +" "+ mc.getNodes() +" "+ cpuTime +" true");
	System.out.println(mc.getSolution());
    }
}
